#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <curses.h>
#include <time.h>

#include "opts.h"
#include "version.h"

#define MAX_ROBOTS 250
/* 250 is PLENTY */

#define WORKX (COLS - 17)
#define WORKY (LINES - 3)
enum pattern { P_DUMB, P_SMART_X, P_SMART_Y, P_RAND, P_MAX };
typedef struct {
	int x, y;
	enum pattern p;
} robot;

struct {
	unsigned int smart:1;
	unsigned int verbose:1;
	unsigned int invincible:2;
} flags;

enum thingie { T_NULL, T_ROBOT, T_SCRAP, T_PLAYER, T_MEAT, T_MAX };
enum bad {
	NOT_SO_BAD,
	BAD_LOCATION = 0x1,
	BAD_ROBOT = 0x2,
	BAD_MAX
};
char graphchar[T_MAX] = " +*@!";

WINDOW *levelwin, *scorewin, *movewin;
long score = 0;
int play_level = 0, stand;
robot robots[MAX_ROBOTS], player;
int robots_left;
enum thingie **level, **templevel;

int set_me_up(void);
void play_me(void);
int get_player_move(void);
void move_robots(void);
void tear_me_down(void);
void draw_borders(void), draw_score(void), draw_moves(void);
void draw_level(void);
void make_level(void), next_level(void);
void handle_args(char *);
void calc_moves(void);
int tele_chances(void);
void usage(int);
int distance(int, int, int, int);
int rnd(int);
int xfrom(int), yfrom(int);
int bad_square(int, int);

char possible_moves[11] = "thjklyubn.";
char moves[11];

void
usage(int status) {
	puts("usage: robots [-hiqsvV]");
	puts("-h	help [print this message]");
	puts("-i	invincible [teleports are safe]");
	puts("-q	quiet [fewer visual updates]");
	puts("-s	smart [smarter robots - beware!]");
	puts("-v	verbose [more visual updates]");
	exit(status);
}

int
distance(int x1, int y1, int x2, int y2) {
	return (int) sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
}

void
calc_moves(void) {
	int i, c, b;

	moves[0] = 't'; /* teleporting is always an option */
	for (i = 1; (c = possible_moves[i]); ++i) {
		if ((b = bad_square(player.y + yfrom(c), player.x + xfrom(c)))) {
			if (b & BAD_ROBOT)
				moves[i] = '!';
			if (b & BAD_LOCATION)
				moves[i] = '*';
		} else {
			moves[i] = c;
		}
	}
	moves[i] = '\0';
}

int
tele_chances(void) {
	int j, k;
	int good = 0;

	for (j = 1; j < WORKY + 1; ++j) {
		for (k = 1; k < WORKX + 1; ++k) {
			if (!bad_square(j, k))
				++good;
		}
	}
	return (100 * good) / (WORKY * WORKX);
}

int
rnd(int n) {
	static int seeded = 0;

	if (!seeded) {
		seeded = 1;
		srand(time(NULL));
	}
	return (n > 1) ? (rand() % n) : 0;
}

void
make_level(void) {
	int i = 10, j, k;

	for (j = 0; j < WORKY + 2; ++j) {
		for (k = 0; k < WORKX + 2; ++k) {
			level[j][k] = T_NULL;
			templevel[j][k] = T_NULL;
		}
	}

	for (j = 0; j < play_level; ++j) {
		i *= 12;
		i /= 10;
		if (i > MAX_ROBOTS) {
			i = MAX_ROBOTS;
			break;
		}
	}

	robots_left = i;

	for (j = 0; j < i; ++j) {
		do {
			robots[j].x = rnd(WORKX) + 1;
			robots[j].y = rnd(WORKY) + 1;
			if (flags.smart) {
				robots[j].p = rnd(P_MAX);
			} else {
				robots[j].p = P_DUMB;
			}
		} while (level[robots[j].y][robots[j].x]);
		level[robots[j].y][robots[j].x] = T_ROBOT;
	}
	do {
		player.x = rnd(WORKX) + 1;
		player.y = rnd(WORKY) + 1;	
	} while (level[player.y][player.x]);
	level[player.y][player.x] = T_PLAYER;
}

void
draw_borders(void) {
	int i;
	char chars[11] = "t-||-\\//\\.";
	int y[11] = { 0, 0, 1, -1, 0, -1, -1, 1, 1, 0 };
	int x[11] = { 0, -1, 0, 0, 1, -1, 1, -1, 1, 0 };

	wmove(stdscr, 0, 0);
	waddch(stdscr, '+');
	for (i = 0; i < WORKX; ++i)
		waddch(stdscr, '-');
	waddch(stdscr, '+');
	wmove(stdscr, LINES - 2, 0);
	waddch(stdscr, '+');
	for (i = 0; i < WORKX; ++i)
		waddch(stdscr, '-');
	waddch(stdscr, '+');
	for (i = 1; i < LINES - 2; ++i) {
		wmove(stdscr, i, 0);
		waddch(stdscr, '|');
		wmove(stdscr, i, WORKX + 1);
		waddch(stdscr, '|');
	}

	for (i = 1; chars[i]; ++i) {
		wmove(movewin, 5 + y[i], 5 + x[i]);
		waddch(movewin, chars[i]);
	}
	wmove(movewin, 1, 1);
	wprintw(movewin, "Robots %s", GAMES_REV);
	wmove(movewin, 9, 1);
	waddstr(movewin, "*  - unsafe");
	wmove(movewin, 10, 1);
	waddstr(movewin, "t  - teleport");
	wmove(movewin, 12, 1);
	waddstr(movewin, "s  - stand");
	wmove(movewin, 13, 1);
	waddstr(movewin, "^L - redraw");
	wrefresh(stdscr);
	wrefresh(movewin);
}

void
draw_moves(void) {
	int i;
	int y[11] = { 0, 0, 2, -2, 0, -2, -2, 2, 2, 0 };
	int x[11] = { 0, -2, 0, 0, 2, -2, 2, -2, 2, 0 };
	if (stand && !flags.verbose)
		return;
	for (i = 0; moves[i]; ++i) {
		wmove(movewin, 5 + y[i], 5 + x[i]);
		waddch(movewin, moves[i]);
	}
	wmove(movewin, 11, 1);
	wprintw(movewin, "[%d%% safe]  ", tele_chances());
	wrefresh(movewin);
}

void
draw_score(void) {
	if (stand && !flags.verbose)
		return;
	werase(scorewin);
	wprintw(scorewin, "Score: %10d %7s Level: %3d   Robots left: %4d", score, "", play_level, robots_left);
	wrefresh(scorewin);
}

void
draw_level(void) {
	int i, j;

	werase(levelwin);
	for (i = 1; i < WORKY + 1; ++i)
		for (j = 1; j < WORKX + 1; ++j)
			mvwaddch(levelwin, i - 1, j - 1, graphchar[level[i][j]]);
	wrefresh(levelwin);
}

int
set_me_up(void) {
	int i;
	initscr();
	level = calloc(LINES - 1, sizeof(enum thingie *));
	templevel = calloc(LINES - 1, sizeof(enum thingie *));
	for (i = 0; i < LINES - 1; ++i) {
		level[i] = calloc(COLS, sizeof(enum thingie));
		templevel[i] = calloc(COLS, sizeof(enum thingie));
	}
	levelwin = newwin(WORKY, WORKX, 1, 1);
	scorewin = newwin(1, COLS, LINES - 1, 0);
	movewin = newwin(WORKY, 15, 0, WORKX + 2);
	nonl();
	cbreak();
	noecho();
	wclear(stdscr);
	for (i = 0; i < MAX_ROBOTS; ++i)
		robots[i].x = robots[i].y = -1;
	draw_borders();
	make_level();
	draw_level();
	draw_score();
	draw_moves();
	return 1;
}

void
next_level(void) {
	++play_level;
	stand = 0;
	make_level();
	draw_level();
	calc_moves();
	draw_moves();
	draw_score();
}

void
tear_me_down(void) {
	int i;
	if (templevel) {
		for (i = 0; i < WORKY + 2; ++i)
			if (templevel[i])
				free(templevel[i]);
		free(templevel);
	}
	if (level) {
		for (i = 0; i < WORKY + 2; ++i)
			if (level[i])
				free(level[i]);
		free(level);
	}
	wclear(stdscr);
	wmove(stdscr, LINES - 5, 0);
	wprintw(stdscr, "You %s on level %d, with %d points.  Better luck next time!\n",
		player.y == -1 ? "quit" : "died", play_level, score);
	wrefresh(stdscr);
	getchar();
	wprintw(stdscr, "\n...and there WILL be a next time.\n");
	wrefresh(stdscr);
	getchar();
	endwin();
	exit(0);
}

int
bad_square(int y, int x){
	int i, j, ret = 0;

	if (level[y][x] && level[y][x] != T_PLAYER)
		ret |= BAD_LOCATION;
	for (i = y - 1; i < y + 2; ++i)
		for (j = x - 1; j < x + 2; ++j)
			if (i >= 0 && j >= 0 && i <= WORKY && j <= WORKX && level[i][j] == T_ROBOT)
				ret |= BAD_ROBOT;
	if (y < 1 || y > WORKY || x < 1 || x > WORKX)
		ret |= BAD_LOCATION;
	return ret;
}

int
xfrom(int c) {
	switch (c) {
		case 'y': case 'h': case 'b':
			return -1;
		case 'u': case 'l': case 'n':
			return 1;
		default:
			return 0;
	}
}

int
yfrom(int c) {
	switch (c) {
		case 'y': case 'k': case 'u':
			return -1;
		case 'b': case 'j': case 'n':
			return 1;
		default:
			return 0;
	}
}

int
get_player_move() {
	int c;
	int y, x, temp;

	while (1) {
		c = stand ? '.' : getchar();

		switch (c) {
			case 'v':
				flags.verbose = !flags.verbose;
				return 0;
				break;
			case 'q':
				player.x = -1;
				player.y = -1;
				return 0;
				break; /* wow, a null statement */
			case 'h': case 'j': case 'k': case 'l':
			case 'y': case 'u': case 'b': case 'n':
			case '.':
				y = player.y + yfrom(c);
				x = player.x + xfrom(c);
				if (bad_square(y, x)) {
					if (!stand) {
						return 0;
					} else
						return 1;
				} else {
					level[player.y][player.x] = T_NULL;
					player.y = y;
					player.x = x;
					level[player.y][player.x] = T_PLAYER;
					return 1;
				}
				break;
			case 's':
				stand = 1;
				break;
			case 't':
				level[player.y][player.x] = T_NULL;
				do {
					player.x = rnd(WORKX) + 1;
					player.y = rnd(WORKY) + 1;	
					temp = bad_square(player.y, player.x);
					if (!flags.invincible) {
						temp &= ~BAD_ROBOT;
					}

				} while (temp);
				level[player.y][player.x] = T_PLAYER;
				return 1;
			case 'L' & 0x1f:
			case 'R' & 0x1f:
				draw_borders();
				draw_level();
				draw_score();
				draw_moves();
				return 0;
				break;
			default:
				return 0;
		}
	}
}

void
move_robots(void) {
	int i, j;
	int x, y;

	for (i = 0; i < WORKY + 2; ++i)
		for (j = 0; j < WORKX + 2; ++j)
			templevel[i][j] =
				((level[i][j] == T_SCRAP) ? T_SCRAP : T_NULL);
	
	templevel[player.y][player.x] = T_PLAYER;

	for (i = 0; i < robots_left; ++i) {
		y = robots[i].y;
		x = robots[i].x;
		if (robots[i].y < player.y)
			y = robots[i].y + 1;
		if (robots[i].y > player.y)
			y = robots[i].y - 1;
		if (robots[i].x < player.x)
			x = robots[i].x + 1;
		if (robots[i].x > player.x)
			x = robots[i].x - 1;

		if (robots[i].p == P_SMART_X) {
			if ((templevel[robots[i].y][x] == T_SCRAP) &&
			    player.y != robots[i].y) {
			    x = robots[i].x;
			    robots[i].p = P_SMART_Y;
			} else if (robots[i].x == player.x)
				robots[i].p = P_SMART_Y;
			else
				y = robots[i].y;
		} else if (robots[i].p == P_SMART_Y) {
			if ((templevel[y][robots[i].x] == T_SCRAP) &&
			    player.x != robots[i].x) {
			    y = robots[i].y;
			    robots[i].p = P_SMART_X;
			} else if (robots[i].y == player.y)
				robots[i].p = P_SMART_X;
			else
				x = robots[i].x;
		} else if (robots[i].p == P_RAND && !rnd(5)) {
			x = robots[i].x + rnd(3) - 1;
			y = robots[i].y + rnd(3) - 1;
		}
		if (robots[i].p != P_DUMB &&
			distance(x, y, player.x, player.y) > 8 && !rnd(10)) {
			do {
				y = rnd(WORKY) + 1;
				x = rnd(WORKX) + 1;
			} while (templevel[y][x] ||
				(distance(x, y, player.x, player.y) > 10));
		}

		y = y > 1 ? y : 1;
		y = y < WORKY ? y : WORKY;
		x = x > 1 ? x : 1;
		x = x < WORKX ? x : WORKX;
		switch (templevel[y][x]) {
			case T_NULL: default:
				templevel[y][x] = T_ROBOT;
				robots[i].y = y;
				robots[i].x = x;
				break;
			case T_SCRAP:
				robots[i].y = -1;
				robots[i].x = -1;
				score += stand ? 20 : 10;
				--robots_left;
				draw_score();
				for (j = i; j < robots_left; ++j) {
					robots[j].y = robots[j + 1].y;
					robots[j].x = robots[j + 1].x;
				}
				--i;
				break;
			case T_ROBOT:
				templevel[y][x] = T_SCRAP;
				score += stand ? 40 : 20;
				for (j = 0; j < i; ++j) {
					if (robots[j].y == y && robots[j].x == x) {
						break;
					}
				}
				--robots_left;
				for (; j < robots_left; ++j) {
					robots[j].y = robots[j + 1].y;
					robots[j].x = robots[j + 1].x;
				}
				--i;
				--robots_left;
				draw_score();
				for (j = i; j < robots_left; ++j) {
					robots[j].y = robots[j + 1].y;
					robots[j].x = robots[j + 1].x;
				}
				--i;
				break;
			case T_PLAYER:
				robots[i].y = -1;
				robots[i].x = -1;
				if (flags.invincible < 2)
					player.x = -1;
				score += stand ? 100 : 10;
				draw_score();
				--robots_left;
				for (j = i; j < robots_left; ++j) {
					robots[j].y = robots[j + 1].y;
					robots[j].x = robots[j + 1].x;
				}
				if (flags.invincible < 2)
					templevel[y][x] = T_MEAT;
				--i;
				break;
		}
	}

	for (i = 0; i < WORKY + 2; ++i)
		for (j = 0; j < WORKX + 2; ++j)
			level[i][j] = templevel[i][j];
}

void
play_me(void) {
	do {
		draw_level();
		calc_moves();
		draw_moves();
		draw_score();
		wmove(levelwin, player.y - 1, player.x - 1);
		wrefresh(levelwin);
		if (get_player_move())
			move_robots();
		if (!robots_left)
			next_level();
	} while(player.x != -1);
	draw_level();
	werase(scorewin);
	wprintw(scorewin, "Munch!  You're robot food! [with %d points]", score);
	wrefresh(scorewin);
	getchar();
	return;
}

int
main(int argc, char *argv[]) {
	int o;

	while ((o = optsopt(argc, argv, "hil#qsvV")) != -1) {
		switch (o) {
		case 'h':
			usage(EXIT_SUCCESS);
			break;
		case 'i':
			flags.invincible = 1;
			break;
		case 'l':
			play_level = strtol(optsarg, 0, 10);
			break;
		case 'q':
			flags.verbose = 0;
			break;
		case 's':
			flags.smart = 1;
			break;
		case 'v':
			flags.verbose = 1;
			break;
		case 'V':
			version();
			break;
		case '?':
			usage(EXIT_FAILURE);
			break;
		}
	}
								 
	if (set_me_up())
		play_me();
	tear_me_down();

	return 0;
}
